# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: custom_types/field_types.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1e\x63ustom_types/field_types.proto\x12\x0c\x63ustom_types\"|\n\x0eMessageWithMap\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x32\n\x03map\x18\x02 \x03(\x0b\x32%.custom_types.MessageWithMap.MapEntry\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"]\n\x13MessageWithOptional\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x0c\n\x04text\x18\x02 \x01(\t\x12\x1a\n\roptional_text\x18\x03 \x01(\tH\x00\x88\x01\x01\x42\x10\n\x0e_optional_text\"1\n\x13MessageWithRepeated\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x0e\n\x06values\x18\x02 \x03(\x05\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'custom_types.field_types_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MESSAGEWITHMAP_MAPENTRY._options = None
  _MESSAGEWITHMAP_MAPENTRY._serialized_options = b'8\001'
  _globals['_MESSAGEWITHMAP']._serialized_start=48
  _globals['_MESSAGEWITHMAP']._serialized_end=172
  _globals['_MESSAGEWITHMAP_MAPENTRY']._serialized_start=130
  _globals['_MESSAGEWITHMAP_MAPENTRY']._serialized_end=172
  _globals['_MESSAGEWITHOPTIONAL']._serialized_start=174
  _globals['_MESSAGEWITHOPTIONAL']._serialized_end=267
  _globals['_MESSAGEWITHREPEATED']._serialized_start=269
  _globals['_MESSAGEWITHREPEATED']._serialized_end=318
# @@protoc_insertion_point(module_scope)
